#ifndef DNest4_DNest4
#define DNest4_DNest4

/*
* Include every header file in DNest4
*/

#include "Version.h"
#include "Barrier.h"
#include "CommandLineOptions.h"
#include "Level.h"
#include "LikelihoodType.h"
#include "Options.h"
#include "RNG.h"
#include "Sampler.h"
#include "Start.h"
#include "Utils.h"
#include "RJObject/RJObject.h"
#include "RJObject/Normals.h"
#include "RJObject/ConditionalPriors/BasicCircular.h"
#include "RJObject/ConditionalPriors/ClassicMassInf1D.h"
#include "RJObject/ConditionalPriors/ConditionalPrior.h"
#include "Distributions/Gaussian.h"
#include "Distributions/Cauchy.h"
#include "Distributions/Laplace.h"
#include "Distributions/Pareto.h"
#include "Distributions/Rayleigh.h"
#include "Distributions/Uniform.h"
#include "Distributions/Exponential.h"
#include "Distributions/LogUniform.h"
#include "Distributions/Jeffreys.h"
#include "Distributions/Kumaraswamy.h"
#include "Distributions/Triangular.h"
#include "Distributions/ContinuousDistribution.h"

#endif

